/*
 * bluCarRc.c
 *
 * Small RC car control program. Receive serial
 * bytes, and set motor power and direction,
 * accordingly. Timeout on missing serial byte
 * will set outputs to zero.
 * 
 * Created: 20.05.2019
 * Author : Eirik Taylor
 */ 

// Includes
#include <avr/io.h>
#include <avr/interrupt.h>
#include "usart_lib.h"
#include "io_lib.h"

// Function declarations
void readCommunicationbuffer(uint8_t temp);

// RC Car states
volatile uint8_t motorPower;
volatile uint8_t motorDirection;
volatile uint8_t steeringAngle;
volatile uint8_t headLightIntensity;
volatile uint8_t brakeLightIntensity;

// Communication state
volatile uint8_t watchDogCounter;

ISR(USART_RX_vect)
{
	bufferInsert(UDR);
}

ISR(TIMER1_OVF_vect)
{
	// 20ms period for the servo updates
	watchDogCounter++;
}

int main(void)
{
	motorPower = 0;
	motorDirection = 0;
	steeringAngle = 0;
	headLightIntensity = 0;
	brakeLightIntensity = 0;
	
	watchDogCounter = 0;
	
	IO_init();
	USART0_init();
	
	// Enable global interrupts
	sei();
	
    while (1) 
    {
		// Check for new data in the USART buffer
		if (writeIndex != currentIndex)
		{
			readCommunicationbuffer(popFromBuffer());
		}
		
		// Ensure that a message packet has been received in the last second
		// otherwise reset all outputs. A new packet should arrive every 100ms or so
		if (watchDogCounter > 20)
		{
			setMotorPowerAndDirection(0, 0);
			setSteeringAngle(127);
			setHeadLights(0);
			setBrakeLights(0);
			
			// Reset it here, so the default settings aren't constantly being reapplied
			watchDogCounter = 0;
		}
    }
}

// Interpret a byte from the USART buffer
void readCommunicationbuffer(uint8_t temp)
{
	// Set commands are any with the MSB set (between 255 and 128)
	// Read commands are any with MSB low, and MSB-1 set (between 127 and 64)
	enum CommandBytes
	{
		SET_MOTOR_POWER = 128,
		GET_MOTOR_POWER = 64,
		SET_MOTOR_DIRECTION = 129,
		GET_MOTOR_DIRECTION = 65,
		SET_STEERING_ANGLE = 130,
		GET_STEERING_ANGLE = 66,
		SET_FRONT_LIGHTS = 131,
		GET_FRONT_LIGHTS = 67,
		SET_BRAKE_LIGHTS = 132,
		GET_BRAKE_LIGHTS = 68
	};
	
	static uint8_t byteCnt;
	static uint8_t commandState;
	
	switch(byteCnt)
	{
		case 0:
		{
			// Perhaps a command byte, try it
			if (temp & 0b11000000)
			{
				byteCnt++;
				commandState = temp;
			}
			else
			{
				byteCnt = 0;
				commandState = 0;
			}
			break;
		}
		
		case 1:
		{
			if (commandState & 0b10000000)
			{
				// A set command, set the correct internal data
				// Reset the watchdog counter for each successful set
				// command.
				switch(commandState)
				{
					case SET_MOTOR_POWER:
					{
						motorPower = temp;
						setMotorPowerAndDirection(motorPower, motorDirection);
						watchDogCounter = 0;
						break;
					}
					
					case SET_MOTOR_DIRECTION:
					{
						motorDirection = temp;
						setMotorPowerAndDirection(motorPower, motorDirection);
						watchDogCounter = 0;
						break;
					}
					
					case SET_STEERING_ANGLE:
					{
						steeringAngle = temp;
						setSteeringAngle(steeringAngle);
						watchDogCounter = 0;
						break;
					}
					
					case SET_FRONT_LIGHTS:
					{
						headLightIntensity = temp;
						setHeadLights(headLightIntensity);
						watchDogCounter = 0;
						break;
					}
					
					case SET_BRAKE_LIGHTS:
					{
						brakeLightIntensity = temp;
						setBrakeLights(brakeLightIntensity);
						watchDogCounter = 0;
						break;
					}
				}
			}
			else
			{
				// A get command, perform the expected action
				switch(commandState)
				{
					case GET_MOTOR_POWER:
					{
						// Transmit the current electro-magnet setting over UART
						USART0_send_byte(GET_MOTOR_POWER);
						USART0_send_byte(motorPower);
						break;
					}
					
					case GET_MOTOR_DIRECTION:
					{
						// Transmit the current electro-magnet setting over UART
						USART0_send_byte(GET_MOTOR_DIRECTION);
						USART0_send_byte(motorDirection);
						break;
					}
					
					case GET_STEERING_ANGLE:
					{
						// Transmit the current electro-magnet setting over UART
						USART0_send_byte(GET_STEERING_ANGLE);
						USART0_send_byte(steeringAngle);
						break;
					}
					
					case GET_FRONT_LIGHTS:
					{
						// Transmit the current electro-magnet setting over UART
						USART0_send_byte(GET_FRONT_LIGHTS);
						USART0_send_byte(headLightIntensity);
						break;
					}
					
					case GET_BRAKE_LIGHTS:
					{
						// Transmit the current electro-magnet setting over UART
						USART0_send_byte(GET_BRAKE_LIGHTS);
						USART0_send_byte(brakeLightIntensity);
						break;
					}
				}
			}
			
			byteCnt = 0;
			commandState = 0;
			break;
		}
		
		default:
		{
			byteCnt = 0;
			commandState = 0;
		}
	}
}
